from functools import partial
from copy import deepcopy
from aga import test_case, test_cases
import ast
import inspect
from contextlib import redirect_stdout
from aga.util import text_diff
from io import StringIO

t_case = partial(test_case, aga_value=0, aga_weight=0)

def check_recursive_ast(fn):
    tree = ast.parse(inspect.getsource(fn))
    for node in ast.walk(tree):
        if isinstance(node, ast.Call):
            if isinstance(node.func, ast.Name):
                if node.func.id == fn.__name__:
                    return True
    return False


def recursion_check(test_case, golden, student):
    
    golden_io  = StringIO()
    student_io = StringIO()

    with redirect_stdout(golden_io):
        golden_result = golden(*deepcopy(test_case.args),
                               **deepcopy(test_case.kwargs))
    with redirect_stdout(student_io):
        student_result = student(*deepcopy(test_case.args),
                                 **deepcopy(test_case.kwargs))

    test_case.assertEqual(golden_result, student_result,
        f"The result returned by {student.__name__} was {repr(student_result)}.\n "
        f"We were expecting a return value of {repr(golden_result)} instead.")
        
    golden_output = golden_io.getvalue()
    student_output = student_io.getvalue()
    
    test_case.assertEqual(golden_output, student_output,
        "Your function's output to the console differs from what was expected. \n"
        "Here are the differences we found:\n"
        f"{text_diff(golden_output, student_output)}")

    test_case.assertTrue(check_recursive_ast(student),
                         f"We've determined that {student.__name__} is not recursive. \n"
                         "Please use recursion (without lambdas) to solve this problem.")


# This is for word_mix problem (we don't care about order)
def recursion_check_sorted(test_case, golden, student):
    
    golden_io  = StringIO()
    student_io = StringIO()

    with redirect_stdout(golden_io):
        golden_result = golden(*deepcopy(test_case.args),
                               **deepcopy(test_case.kwargs)).sort()
    with redirect_stdout(student_io):
        student_result = student(*deepcopy(test_case.args),
                                 **deepcopy(test_case.kwargs)).sort()

    test_case.assertEqual(golden_result, student_result,
        f"The result returned by {student.__name__} was {repr(student_result)}.\n "
        f"We were expecting a return value of {repr(golden_result)} instead.")
        
    golden_output = golden_io.getvalue()
    student_output = student_io.getvalue()
    
    test_case.assertEqual(golden_output, student_output,
        "Your function's output to the console differs from what was expected. \n"
        "Here are the differences we found:\n"
        f"{text_diff(golden_output, student_output)}")

    test_case.assertTrue(check_recursive_ast(student),
                         f"We've determined that {student.__name__} is not recursive. \n"
                         "Please use recursion to solve this problem.")


uses_recursion_t_case = partial(test_case, aga_override_test=recursion_check, aga_value=0, aga_weight=0)
uses_recursion_t_case_sorted = partial(test_case, aga_override_test=recursion_check_sorted, aga_value=0, aga_weight=0)
uses_recursion_t_cases = partial(test_cases.zip, aga_override_test=recursion_check, aga_value=0, aga_weight=0)
uses_recursion_t_cases_sorted = partial(test_cases.zip, aga_override_test=recursion_check_sorted, aga_value=0, aga_weight=0)


def award_prize(tests, metadata):
    """
    Awards a prize (points for the problem) depending on how many submissions it took to succeed and if it's late.
    * If not passing all tests, give no points.
    * Else, if late, give half credit.
    * Else (thus on-time and correct), give 1 less point per failed submission (assuming that it's worth 20 points).
    """
    on_time = metadata.is_on_time()
    all_correct = all(t.is_correct() for t in tests)

    # If wrong...
    if not all_correct:
        # ...encourage them to fix everything.
        return (0.0, "To earn these points, make sure all the tests pass!")

    # If correct but late...
    if not on_time:
        return (0.5, "Great work making all the tests pass! You get 10 points total for a late submission.")
    
    # If correct and on time...
    # ...determine the fraction of the prize based on the number of attempts.
    failed_attempts = metadata.previous_submissions
    if failed_attempts >= 7:
        return (0.65,
                "Great work passing all the tests! "
                f"You submitted {failed_attempts+1} times, "
                "so you earn the minimum of 13 points. "
                "Nice job sticking with it! "
                "But please try as much as possible "
                "to test your code before submitting it.")
    else:
        suffix = ["st", "nd", "rd", "th"][failed_attempts if failed_attempts <= 2 else 3]
        points = 1.0 - (failed_attempts * 0.05)
        if failed_attempts == 0:
            return (points,
                    "Great work passing all the tests! This is your first attempt, so you earn the full 20 points. Hurray!")
        return (points,
               "Great work passing all the tests! This is your "
                f"{failed_attempts+1}{suffix} attempt, so "
                f"you earn {failed_attempts} points fewer, which is {20 - failed_attempts} in total.")
